"""Implement the server side of Socket.IO.

    https://github.com/learnboost/socket.io-spec

Ah, abstraction! This is a whole convoluted mess to provide some pretty nice
API inside socket resources. Here are the objects involved on the server side,
from the inside out:

    Message     a Socket.IO message, a colon-delimited set of bytestrings
    Packet      a Socket.IO packet, a message or series of framed messages
    Buffer      a Socket.IO buffer, buffers incoming and outgoing messages
    Loop        an object responsible for repeatedly calling socket.tick
    Socket      a Socket.IO socket, maintains state
    Channel     an object that represents all connections to a single Resource
    Transport   a Socket.IO transport mechanism, does HTTP work
    Resource    an HTTP resource, a file on your filesystem, application logic
    Response    an HTTP Response message
    Request     an HTTP Request message

    Engine      fits somewhere, handles networking implementation; Buffer and
                  Loop attributes point to implementations of the above


A specially-crafted HTTP request creates a new Socket. That socket object
exists until one of these conditions is met:

    - the application explicitly disconnects
    - the client explicitly disconnects
    - the client disappears (for some definition of "disappears")

A second specially-crafted HTTP request negotiates a Transport. Subsequent
specially-crafted HTTP requests are marshalled into socket reads and writes
according to the Transport negotiated.

The Loop object is responsible for running socket.tick until it is told to stop
(as a result of one of the above three conditions). socket.tick exec's the
third page of the application's socket resource in question. This code is
expected to block. For ThreadedLoop that means we can't stop the loop inside of
native code. The ThreadedBuffer object cooperates with ThreadedLoop, so if your
application only ever blocks on socket.recv then you are okay. CooperativeLoops
should be immediately terminable assuming your application and its dependencies
cooperate ;-).

"""
from aspen import Response

FFFD = u'\ufffd'.encode('utf-8')
HEARTBEAT = 15
TIMEOUT = 10
TRANSPORTS = ['xhr-polling']

from aspen.sockets.channel import Channel
from aspen.sockets.socket import Socket
from aspen.sockets.transport import XHRPollingTransport


__sockets__ = {}
__channels__ = {}


def get(request):
    """Takes a Request object and returns a Response or Transport object.

    When we get the request it has socket set to a string, the path part after
    *.sock, which is something like 1/websocket/43ef6fe7?foo=bar.

        1           protocol (we only support 1)
        websocket   transport
        43ef6fe7    socket id (sid)
        ?foo=bar    querystring

    The Socket.IO handshake is a GET request to 1/. We return Response for the
    handshake. After the handshake, subsequent messages are to the full URL as
    above. We return a Transported instance for actual messages.

    """

    # Exit early.
    # ===========

    if request.socket is None:
        return None


    # Parse and validate the socket URL.
    # ==================================

    parts = request.socket.split('/')
    nparts = len(parts)
    if nparts not in (2, 3):
        msg = "Expected 2 or 3 path parts for Socket.IO socket, got %d."
        raise Response(400, msg % nparts)

    protocol = parts[0]
    if protocol != '1':
        msg = "Expected Socket.IO protocol version 1, got %s."
        raise Response(400, msg % protocol)


    # Handshake
    # =========

    if len(parts) == 2:

        # Note that since we're indexing on URI path instead of fs path,
        # wildcard socket simplates like %foo.sock will end up differentiated
        # here.

        path = request.line.uri.path.raw
        if path in __channels__:
            channel = __channels__[path]
        else:
            channel = Channel(path, request.website.network_engine.Buffer)
            __channels__[path] = channel

        socket = Socket(request, channel)
        assert socket.sid not in __sockets__ # sanity check
        __sockets__[socket.sid] = socket
        socket.loop.start()

        return socket.shake_hands() # a Response


    # More than a handshake.
    # ======================

    transport = parts[1]
    sid = parts[2]

    if transport not in TRANSPORTS:
        msg = "Expected transport in {%s}, got %s."
        msg %= (",".join(TRANSPORTS), transport)
        raise Response(400, msg)

    if sid not in __sockets__:
        msg = "Expected %s in cache, didn't find it"
        raise Response(400, msg % sid)

    if type(__sockets__[sid]) is Socket:
        # This is the first request after a handshake. It's not until this
        # point that we know what transport the client wants to use.
        Transport = XHRPollingTransport # XXX derp
        __sockets__[sid] = Transport(__sockets__[sid])

    transport = __sockets__[sid]
    return transport



from Cookie import SimpleCookie
from StringIO import StringIO

from aspen.http.request import Request
from aspen.testing import StubWSGIRequest

BOUNDARY = 'BoUnDaRyStRiNg'
MULTIPART_CONTENT = 'multipart/form-data; boundary=%s' % BOUNDARY


def encode_multipart(boundary, data):
    """
    Encodes multipart POST data from a dictionary of form values.

    Borrowed from Django
    The key will be used as the form data name; the value will be transmitted
    as content. If the value is a file, the contents of the file will be sent
    as an application/octet-stream; otherwise, str(value) will be sent.
    """
    lines = []

    for (key, value) in data.items():
        lines.extend([
            '--' + boundary,
            'Content-Disposition: form-data; name="%s"' % str(key),
            '',
            str(value)
        ])

    lines.extend([
        '--' + boundary + '--',
        '',
    ])
    return '\r\n'.join(lines)


class TestClient(object):
    """
    The Aspen test client.

    Used in tests to emulate ``GET`` and ``POST`` requests by sending them
    into a ``Website`` instance's ``handle_safely`` method.

    Aspen does not define any User data structures or modules. If you want to
    do anything with users/sessions etc in your tests it is expected that you
    will subclass this TestClient and add a ``add_cookie_info`` method.

    For example, in gittip a suitable subclass might be::

        class GittipTestClient(TestClient):

            def add_cookie_info(self, request, cookie_info):
                if cookie_info:
                    user = cookie_info.get('user')
                    if user is not None:
                        user = User.from_id(user)
                        # Note that Cookie needs a bytestring.
                        request.headers.cookie['session'] = user.session_token

    Example usage in a test::

        def test_api_handles_posts():
            client = TestClient(website)

            # We need to get ourselves a token!
            response = client.get('/')
            csrf_token = response.request.context['csrf_token']

            # Then, add a $1.50 and $3.00 tip
            response = client.post("/test_tippee1/tip.json",
                                {'amount': "1.00", 'csrf_token': csrf_token},
                                cookie_info={'user': 'test_tipper'})

            # Confirm we get back the right amounts in the JSON body.
            first_data = json.loads(response.body)
            assert_equal(first_data['amount'], "1.00")
    """
    def __init__(self, website):
        self.cookies = SimpleCookie()
        self.test_website = website

    def get_request(self, path, method="GET", body=None,
                    **extra):
        env = StubWSGIRequest(path)
        env['REQUEST_METHOD'] = method
        env['wsgi.input'] = StringIO(body)
        env['HTTP_COOKIE'] = self.cookies.output(header='', sep='; ')
        env.update(extra)
        return Request.from_wsgi(env)

    def perform_request(self, request, cookie_info):
        request.website = self.test_website
        self.add_cookie_info(request, **(cookie_info or {}))
        response = self.test_website.handle_safely(request)
        if response.headers.cookie:
            self.cookies.update(response.headers.cookie)
        return response

    def add_cookie_info(self, request, **cookie_info):
        """Place holder function that can be replaced in a subclass.

        For example in gittip.com, it might be of interest to load session
        information into the cookie like this::

            if cookie_info:
                user = cookie_info.get('user')
                if user is not None:
                    user = User.from_id(user)
                    # Note that Cookie needs a bytestring.
                    request.headers.cookie['session'] = user.session_token
        """
        pass

    def post(self, path, data, content_type=MULTIPART_CONTENT,
             cookie_info=None, **extra):
        """Perform a dummy POST request against the test website.

        :param path:
            The url to perform the virutal-POST to.

        :param data:
            A dictionary or list of tuples to be encoded before being POSTed.

        Any additional parameters will be sent as headers. NOTE that in Aspen
        (request.py make_franken_headers) only headers beginning with ``HTTP``
        are included in the request - and those are changed to no longer
        include ``HTTP``. There are currently 2 exceptions to this:
        ``'CONTENT_TYPE'``, ``'CONTENT_LENGTH'`` which are explicitly checked
        for.
        """
        post_data = data

        if content_type is MULTIPART_CONTENT:
            post_data = encode_multipart(BOUNDARY, data)

        request = self.get_request(path, "POST", post_data,
                                   CONTENT_TYPE=str(content_type),
                                   **extra)
        return self.perform_request(request, cookie_info)

    def get(self, path, cookie_info=None, **extra):
        request = self.get_request(path, "GET")
        return self.perform_request(request, cookie_info)



from aspen import json, Response


RESERVED_EVENTS = [ 'message'
                  , 'connect'
                  , 'disconnect'
                  , 'open'
                  , 'close'
                  , 'error'
                  , 'retry'
                  , 'reconnect'
                   ]


class Message(object):
    """Model a Socket.IO message.
    """

    def __init__(self, type_=0, id='', endpoint='', data=''):
        self.type = type_
        self.id = id
        self.endpoint = endpoint
        self.data = data

    @classmethod
    def from_bytes(cls, bytes):
        parts = bytes.split(':', 3)
        if len(parts) == 3:
            parts.append('') # data part is optional
        if len(parts) != 4: # "::".split(":", 3) == ['', '', '']
            raise SyntaxError("This message has too few colons: %s." % bytes)
        return cls(*parts)

    def __repr__(self):
        return "<Message %s>" % self

    def __str__(self):
        data = self.data
        if self.type in (4, 5):
            data = json.dumps(data)
        return ":".join([ str(self.type)
                        , self.id
                        , self.endpoint
                        , data
                         ])

    def __cmp__(self, other):
        return cmp(str(self), str(other))

    # type
    # ====

    __type = 0

    def _get_type(self):
        return self.__type

    def _set_type(self, type_):
        try:
            type_ = int(type_)
            assert type_ in range(9)
        except (ValueError, AssertionError), exc:
            raise ValueError("The message type is not in 0..8: %s." % type_)
        self.__type = type_

    type = property(_get_type, _set_type)


    # data
    # ====

    __data = ''

    def _get_data(self):
        return self.__data

    def _set_data(self, data):
        if self.type == 4:              # json
            data = json.loads(data)
        elif self.type == 5:            # event
            data = json.loads(data)
            if 'name' not in data:
                raise ValueError("An event message must have a 'name' key.")
            if 'args' not in data:
                raise ValueError("An event message must have an 'args' key.")
            if data['name'] in RESERVED_EVENTS:
                msg = "That event name is reserved: %s." % data['name']
                raise ValueError(msg)
        self.__data = data

    data = property(_get_data, _set_data)




import re


SPLITTER = '^\[---+\](?P<header>.*?)(\n|$)'
ESCAPED_SPLITTER = '^\\\\(\\\\*)(\[---+\].*?(\n|$))'
SPECLINE_SPLIT = '(?:\s+|^)via\s+'

SPLITTER = re.compile(SPLITTER, re.MULTILINE)
ESCAPED_SPLITTER = re.compile(ESCAPED_SPLITTER, re.MULTILINE)
SPECLINE_SPLIT = re.compile(SPECLINE_SPLIT)


class Page(object):
    __slots__ = ('header', 'content', 'offset')

    def __init__(self, content, header='', offset=0):
        self.content = content
        self.header = header
        self.offset = offset

    @property
    def padded_content(self):
        return ('\n' * self.offset) + self.content


def split(raw):
    '''Pure split generator. This function defines the plain logic to split a
    string into a list of pages
    '''

    current_index = 0
    line_offset = 0
    header = ''

    for page_break in SPLITTER.finditer(raw):
        content = raw[current_index:page_break.start()]
        yield Page(content, header, line_offset)
        line_offset += content.count('\n') + 1
        header = page_break.group('header').strip()
        current_index = page_break.end()

    # Yield final page. If no page dividers were found, this will be all of it
    content = raw[current_index:]
    yield Page(content, header, line_offset)

def escape(content):
    '''Pure escape method. This function defines the logic to properly convert
    escaped splitter patterns in a string
    '''
    return ESCAPED_SPLITTER.sub(r'\1\2', content)

def split_and_escape(raw):
    '''This function defines the logic to split and escape a string.
    '''
    for page in split(raw):
        page.content = escape(page.content)
        yield page

def parse_specline(header):
    '''Attempt to parse the header in a page returned from split(...) as a
    specline. Returns a tuple (content_type, renderer)
    '''
    parts = SPECLINE_SPLIT.split(header, 1) + ['']
    return parts[0].strip(), parts[1].strip()

def can_split(raw, splitter=SPLITTER):
    '''Determine if a text block would be split by a splitter
    '''
    return bool(SPLITTER.search(raw))





from aspen import Response

def hook(request):
    """A hook to return 200 to an 'OPTIONS *' request"""
    if request.line.method == "OPTIONS" and request.line.uri == "*":
        raise Response(200)
    return request

